class Anomaly_index:
    """
    marker of the detected points
    that anomaly or non-anomaly
    """

    def info_anomalies(self, extracted_or_anomalies, points=bool):
        if points is True:
            for res in extracted_or_anomalies:
                res['anomaly'] = 1
        else:
            for res in extracted_or_anomalies:
                res['anomaly'] = 0
        return extracted_or_anomalies

    def update_info(self, mark, detect=bool, reason=None):
        if detect is True:
            mark['anomaly'] = 1
            mark['anomaly_reason'] = reason
        else:
            mark['anomaly'] = 0
        return mark



import math


class Distance:
    """
    the distance between two GPS points dec calculated
    using latitude and longitude information.
    the distances remaining above the "distance_limit"
    parameter will be marked as the anomaly point.
    """
    distance_limit = 50
    distance_low_limit = 0.8

    def __init__(self):
        self.WGS84_a = 6378137.0
        self.WGS84_b = 6356752.314245

    def ecef_from_lla(self, lat, lon, alt):
        """
        Compute ecef XYZ from latitude, longitude and altitude.

        All using the WGS94 model.
        Altitude is the distance to the WGS94 ellipsoid.
        Check results here https://www.oc.nps.edu/oc2902w/coord/llhxyz.ht
        """
        a2 = self.WGS84_a ** 2
        b2 = self.WGS84_b ** 2
        lat = math.radians(lat)
        lon = math.radians(lon)
        ll = 1.0 / math.sqrt(a2 * math.cos(lat) ** 2 + b2 * math.sin(lat) ** 2)
        x = (a2 * ll + alt) * math.cos(lat) * math.cos(lon)
        y = (a2 * ll + alt) * math.cos(lat) * math.sin(lon)
        z = (b2 * ll + alt) * math.sin(lat)
        return x, y, z

    def gps_distance(self, latlon_1, latlon_2):
        """
        :param latlon_1: 2 Latitude points
        :param latlon_2: 2 Longitude points
        :return:
        """
        x1, y1, z1 = self.ecef_from_lla(float(latlon_1[0]), float(latlon_2[0]), 0.0)
        x2, y2, z2 = self.ecef_from_lla(float(latlon_1[1]), float(latlon_2[1]), 0.0)
        dis = math.sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2 + (z1 - z2) ** 2)
        return dis



class Angle:
    """
    The length of the "heading" values between two
    points gives the difference between the "heading".
    The values that remain above the parameter that,
    "header_limit" will be marked as an anomaly.
    """
    header_limit = 90

    def __init__(self):
        self.upper_angle = int
        self.lower_angle = int

    def rang(self, pair_head: list) -> int:
        self.upper_angle = int(max(pair_head[0], pair_head[1]))
        self.lower_angle = int(min(pair_head[0], pair_head[1]))
        range_angle = range(self.lower_angle, self.upper_angle)
        range_angle = [a % 360 if a >= 360 else a for a in range_angle]
        angle1 = len(range_angle)
        angle2 = (360 - angle1) % 360
        res_angle = min(angle2, angle1)
        return res_angle



from .one_sequence import Sequence
import os
import re
from gps_anomaly.config.distance import Distance


def first_point(anomaly_points, extracted_seq, seq):
    """
    First points of sequences has marked as anomaly or
    not anomaly, with comparing second point.
    """
    if len(extracted_seq) > 1 and seq[1] in extracted_seq:
        extracted_seq.append(seq[0])
    else:
        anomaly_points.append(seq[0])
    return extracted_seq, anomaly_points


class Info:
    """
    Notices failure sequence UUID,
    anomaly points' image names.
    """

    def create_list_filename(self, distrubution):
        """
        appends result to one list
        :param distrubution:
        :return:
        """
        file_list = []
        for sequences in distrubution:
            for seq in sequences:
                if type(seq) == dict:
                    file_list.append(os.path.join(seq['filename']))
        return file_list

    def check_distance_from_desc(self, desc):
        """
        Checks the distance between images, marks as failed or duplicate.
        """
        if 'anomaly_reason' in desc:
            reason = desc['anomaly_reason']
            match = re.search(r'The distance:\s*([-+]?[0-9]*\.?[0-9]+)', reason)
            if match:
                distance = float(match.group(1))
                if distance < Distance.distance_low_limit:
                    return "duplicate"
                else:
                    return "failed"

        return None

    def update_info(self, descs, info, uud):
        for desc in descs:
            if self.check_distance_from_desc(desc) == "duplicate":
                info["Information"]["duplicated_images"] += 1
                info["Information"]["processed_images"] -= 1
            elif self.check_distance_from_desc(desc) == "failed":
                info["Information"]["failed_images"] += 1
                info["Information"]["processed_images"] -= 1
        uud = [lis for lis in uud if lis != []]
        info['Information']['anomaly_sequences'] = uud
        return info

    def create_json(self, distrubution, info):
        """
        appends result to one list
        :param info:
        :param distrubution:
        :return:
        """
        united_sequences = []
        for sequences in distrubution:
            for seq in sequences:
                if type(seq) == dict:
                    united_sequences.append(seq)
        united_sequences.append(info)
        return united_sequences


def mark_points(desc):
    """
        - Group the sequences, remove that has one element sequences.
        - For each series, calculate distance of all points in the sequence.
        - using length anomaly/sequence ratio, get the final result sequences extracted anomalies
    """
    data_all = Sequence(desc)
    info = Info()
    extracted, information, anomaly, uud, with_anomaly, order_seq = data_all.groupy_to_result()
    filenames_list = info.create_list_filename(anomaly)
    information = info.update_info(desc, information, uud)
    order_seq = info.create_json(order_seq, information)
    anomaly_points = info.create_json(anomaly, information)
    return order_seq, filenames_list, anomaly_points



import urllib.request
import json
from gps_anomaly.detector import Anomaly

res = urllib.request.urlopen(
    'https://raw.githubusercontent.com/mapilio/mapilio-kit/master/schema/image_description_schema.json')
res_body = res.read()

# https://docs.python.org/3/library/json.html
json_file = json.loads(res_body.decode("utf-8"))

inforrmation = {"Information":
    {
        "total_images": 7,
        "processed_images": 7,
        "failed_images": 0,
        "duplicated_images": 0,
    }
}

json_file.append(inforrmation)
anomaly = Anomaly()
result, failure_images, _ = anomaly.anomaly_detector(frames=json_file)

with open(file="image_description_schema_export.json", mode="w") as f:
    json.dump(result, f, indent=4)



from typing import List, Any
from itertools import groupby
from .config.heading import Angle
from .config.info_anomalies import Anomaly_index
from .config.distance import Distance
import itertools

DOWN_RATIO = 0.2
SEQUENCE_LIMIT = 10
ALTITUDE_LOWER = 0    ##removed this condition, can be added on line71
ALTITUDE_UPPER = 3500 ## La paz/Bolivia


class AnomalyConfig:
    down_percent: float = DOWN_RATIO
    sequence_limit: int = SEQUENCE_LIMIT
    altitude_lower: int = ALTITUDE_LOWER
    altitude_upper: int = ALTITUDE_UPPER


def pairwise(iterable):
    """s -> (s0,s1), (s1,s2), (s2, s3), ..."""
    a, b = itertools.tee(iterable)
    next(b, None)
    return list(zip(a, b))


def key_func(k):
    return k['sequenceUuid']


class Sequence:
    """
    sequence based anomaly
    detects and returns anomaly points
    """

    extracted_seq: List[Any]
    anomaly_points: List[Any]

    def __init__(self, descs):
        self.information = list(descs).pop()
        descs = list(descs)[:-1]
        self.descs = [desc for desc in descs if ("error" not in desc) and ("heading" in desc)]
        self.sequences = []
        self.distribution = []
        self.anomalies = []
        self.withanomalies = []
        self.uud = []
        self.order_seq = []


    def list_dist(self, zipped, seq, pair_head):
        """
        :param pair_head:
        :param zipped: format (latitude,latitude,longitude,longitude)
        :param seq: squence
        :return: data with anomalies removed
        """
        extracted_seq = []
        anomaly_points = []
        ai = Anomaly_index()
        ang = Angle()
        distance = Distance()
        if len(seq) < AnomalyConfig.sequence_limit:
            anomaly_points = seq
        else:
            for i in range(len(zipped)):
                rang = ang.rang(pair_head[i])
                reasons = []
                dis=distance.gps_distance(zipped[i][0],zipped[i][1])
                # Check distance condition
                if not (distance.distance_low_limit < dis < distance.distance_limit):
                    reasons.append(f"Distance condition not met. The distance: {dis}")

                # Check range condition
                if rang >= ang.header_limit:
                    reasons.append("Range condition not met")

                dicton = next(dic for dic in seq if dic['latitude'] == zipped[i][0][1])

                # Check altitude condition
                if AnomalyConfig.altitude_upper <= dicton['altitude']:
                    reasons.append("Altitude above upper limit")

                # If any reason is present, handle as anomaly
                if reasons:
                    anomaly_points.append(dicton)
                    reason_text = ", ".join(reasons)
                    seq[i] = ai.update_info(seq[i], detect=True, reason=reason_text)
                    if i == len(zipped) - 1:  # Adjusted to avoid index out of range error
                        seq[i] = ai.update_info(seq[i], detect=True, reason=reason_text)
                        seq[i + 1] = ai.update_info(seq[i + 1], detect=True, reason=reason_text)
                else:
                    # If no anomaly, handle normally
                    extracted_seq.append(dicton)
                    seq[i] = ai.update_info(seq[i], detect=False)
                    if i == len(zipped) - 1:  # Adjusted to avoid index out of range error
                        seq[i] = ai.update_info(seq[i], detect=False)
                        seq[i + 1] = ai.update_info(seq[i + 1], detect=False)

        # extracted_seq, anomaly_points = first_point(anomaly_points,extracted_seq,seq)
        ratio_seq = len(extracted_seq) / (len(extracted_seq) + len(anomaly_points))

        if ratio_seq < AnomalyConfig.down_percent or len(extracted_seq) < 5:
            extracted = []
            seq = ai.info_anomalies(seq, True)
            anomalies = seq
            uuud = (anomalies[0]['sequenceUuid'])

        else:
            extracted = extracted_seq
            anomalies = anomaly_points
            uuud = []

        extracted = ai.info_anomalies(extracted, False)
        anomalies = ai.info_anomalies(anomalies, True)
        withanomaly = extracted + anomalies
        # withanomaly.sort(key=lambda x: x['CaptureTime'])

        return extracted, anomalies, uuud, withanomaly, seq

    def groupy_to_result(self):
        """
        - Group the sequences, remove that has one element sequences.
        - For each series, calculate distance of all points in the sequences.
        - using anomaly/sequence ratio, get the final result sequences extracted anomalies
        :return:
        """
        self.descs = sorted(self.descs, key=key_func)
        for _, val in groupby(self.descs, key_func):
            self.sequences.append(list(val))
        for sequen in self.sequences:
            latitude = []
            longitude = []
            heading = []
            for seq in sequen:
                latitude.append(seq['latitude'])
                longitude.append(seq['longitude'])
                heading.append(seq['heading'])
            pair_lat, pair_lon, pair_head = pairwise(latitude), pairwise(longitude), pairwise(heading)
            zipped = list(zip(pair_lat, pair_lon))
            extracted, anomalies, uuud, withanomaly, orderseq = self.list_dist(zipped, sequen, pair_head)
            self.distribution.append(extracted)
            self.anomalies.append(anomalies)
            self.uud.append(uuud)
            self.withanomalies.append(withanomaly)
            self.order_seq.append(orderseq)

        return self.distribution, self.information, self.anomalies, self.uud, self.withanomalies, self.order_seq



import io
import os
import re

import setuptools


def get_long_desc():
    base_dir = os.path.abspath(os.path.dirname(__file__))
    with io.open(os.path.join(base_dir, "README.md"), encoding = "utf-8") as f:
        return f.read()

def get_requirements():
    with open("requirements.txt", encoding = "utf-8") as f:
        return f.read().splitlines()


def get_version():
    cwd = os.path.abspath(os.path.dirname(__file__))
    current_version = os.path.join(cwd, "gps_anomaly", "version.py")
    with io.open(current_version, encoding = "utf-8") as f:
        return re.search(r'^__version__ = [\'"]([^\'"]*)[\'"]', f.read(), re.M).group(1)

setuptools.setup(
    name = "gps-anomaly-detector",
    version = get_version(),
    author = "Mapilio",
    description = "Gps Anomaly Detector of Mapilio",
    long_description = get_long_desc(),
    long_description_content_type='text/markdown',
    url = "https://github.com/mapilio/gps_anomaly_detection",
    license='MIT License',
    python_requires='>=3.7',
    install_requires = get_requirements(),
)



from gps_anomaly.detect_functions import mark_points


class Anomaly:
    """
    Detector anomalies launch class
    """

    @staticmethod
    def anomaly_detector(frames):
        """
        :param frames:
        :return:
        """
        result, failed_imgs, _ = mark_points(frames)

        return result, failed_imgs, _
